﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Collections.Concurrent;
using System.Reflection.Emit;
using System.Threading;
using System.Runtime.CompilerServices;

namespace Dapper.Contrib.Extensions
{

    public static class SqlMapperExtensions
    {
        public interface IProxy
        {
            bool IsDirty { get; set; }
        }
        private static readonly ConcurrentDictionary<RuntimeTypeHandle, string> GetQueries = new ConcurrentDictionary<RuntimeTypeHandle, string>();
        private static readonly ConcurrentDictionary<RuntimeTypeHandle, string> TypeTableName = new ConcurrentDictionary<RuntimeTypeHandle, string>();


        private static readonly ConcurrentDictionary<RuntimeTypeHandle, IEnumerable<MyMemberInfo>> TypeProperties = new ConcurrentDictionary<RuntimeTypeHandle, IEnumerable<MyMemberInfo>>();
        private static IEnumerable<MyMemberInfo> TypePropertiesCache(Type type)
        {
            if (TypeProperties.ContainsKey(type.TypeHandle))
            {
                return TypeProperties[type.TypeHandle];
            }

            var properties = type.GetProperties();
            var fields = type.GetFields();
            IList<MyMemberInfo> list = new List<MyMemberInfo>();
            foreach (var item in properties)
            {
                MyMemberInfo mmi = new MyMemberInfo();
                //Column
                var columnattr = item.GetCustomAttributes(true).FirstOrDefault(a => a is ColumnAttribute);
                if (columnattr != null)
                {
                    mmi.Column = ((ColumnAttribute)columnattr).Name;
                }
                else
                {
                    mmi.Column = item.Name;
                }
                //Key
                var keyattr = item.GetCustomAttributes(true).FirstOrDefault(a => a is KeyAttribute);
                if (keyattr != null)
                {
                    mmi.IsKey = true;
                    mmi.IsKeyIncrement = ((KeyAttribute)keyattr).IsIncrement;
                }
                //Ignore
                var ignoreattr = item.GetCustomAttributes(true).FirstOrDefault(a => a is IgnoreAttribute);
                if (ignoreattr != null)
                {
                    mmi.IsIgnore = true;
                    mmi.IgnoreType = ((IgnoreAttribute)ignoreattr).Type;
                }
                //
                mmi.PropertyInfo = item;
                list.Add(mmi);
            }
            foreach (var item in fields)
            {
                MyMemberInfo mmi = new MyMemberInfo();
                //Column
                var columnattr = item.GetCustomAttributes(true).FirstOrDefault(a => a is ColumnAttribute);
                if (columnattr != null)
                {
                    mmi.Column = ((ColumnAttribute)columnattr).Name;
                }
                else
                {
                    mmi.Column = item.Name;
                }
                //Key
                var keyattr = item.GetCustomAttributes(true).FirstOrDefault(a => a is KeyAttribute);
                if (keyattr != null)
                {
                    mmi.IsKey = true;
                    mmi.IsKeyIncrement = ((KeyAttribute)keyattr).IsIncrement;
                }
                //Ignore
                var ignoreattr = item.GetCustomAttributes(true).FirstOrDefault(a => a is IgnoreAttribute);
                if (ignoreattr != null)
                {
                    mmi.IsIgnore = true;
                    mmi.IgnoreType = ((IgnoreAttribute)ignoreattr).Type;
                }
                //
                mmi.FieldInfo = item;
                list.Add(mmi);
            }

            TypeProperties[type.TypeHandle] = list;
            return list;
        }
        private static string GetTableName(Type type)
        {
            string name;
            if (!TypeTableName.TryGetValue(type.TypeHandle, out name))
            {
                name = type.Name;
                var tableattr = type.GetCustomAttributes(true).FirstOrDefault(a => a is TableAttribute);
                if (tableattr != null)
                    name = ((ColumnAttribute)tableattr).Name;
                TypeTableName[type.TypeHandle] = name;
            }
            return name;
        }




        /// <summary>
        /// Returns a single entity by a single id from table "Ts". T must be of interface type. 
        /// Id must be marked with [Key] attribute.
        /// Created entity is tracked/intercepted for changes and used by the Update() extension. 
        /// </summary>
        /// <typeparam name="T">Interface type to create and populate</typeparam>
        /// <param name="connection">Open SqlConnection</param>
        /// <param name="id">Id of the entity to get, must be marked with [Key] attribute</param>
        /// <returns>Entity of T</returns>
        public static T Get<T>(this IDbConnection connection, object id, IDbTransaction transaction = null, int? commandTimeout = null) where T : class
        {
            var type = typeof(T);
            string sql;
            if (!GetQueries.TryGetValue(type.TypeHandle, out sql))
            {
                var allProperties = TypePropertiesCache(type);
                var keyProperties = allProperties.Where(op => op.IsKey);
                if (keyProperties.Count() > 1)
                    throw new DataException("Get<T> only supports an entity with a single [Key] property");
                if (keyProperties.Count() == 0)
                    throw new DataException("Get<T> only supports en entity with a [Key] property");

                var name = GetTableName(type);

                var sb = new StringBuilder(null);
                sb.Append("select ");
                for (var i = 0; i < allProperties.Count(); i++)
                {
                    var property = allProperties.ElementAt(i);
                    if (property.IsIgnore && property.GetIgnore(IgnoreType.Select)) continue;
                    sb.AppendFormat("{0} as {1}", property.Column, property.Name);
                    sb.Append(", ");
                }
                sb.Remove(sb.Length - 2, 2);
                sb.AppendFormat(" from {0}", name);
                sb.Append(" where ");
                for (var i = 0; i < keyProperties.Count(); i++)
                {
                    var property = keyProperties.ElementAt(i);
                    sb.AppendFormat("{0} = #{1}#", property.Column, property.Name);
                    sb.AppendFormat(" and ");
                }
                sb.Remove(sb.Length - 5, 5);
                sql = sb.ToString();
                GetQueries[type.TypeHandle] = sql;
            }

            T obj = null;

            if (type.IsInterface)
            {
                var res = SqlMapper.Query<Dictionary<string, object>>(connection, sql, id).FirstOrDefault();

                if (res == null)
                    return (T)((object)null);

                obj = ProxyGenerator.GetInterfaceProxy<T>();

                foreach (var property in TypePropertiesCache(type))
                {
                    var val = res[property.Name];
                    property.SetValue(obj, val);
                }

                ((IProxy)obj).IsDirty = false;   //reset change tracking and return
            }
            else
            {
                obj = SqlMapper.Query<T>(connection, sql, id, false, transaction, commandTimeout, null).FirstOrDefault();
            }
            return obj;
        }

        /// <summary>
        /// Inserts an entity into table "Ts" and returns identity id.
        /// </summary>
        /// <param name="connection">Open SqlConnection</param>
        /// <param name="entityToInsert">Entity to insert</param>
        /// <returns>Identity of inserted entity</returns>
        public static bool Insert<T>(this IDbConnection connection, T entityToInsert, IDbTransaction transaction = null, int? commandTimeout = null) where T : class
        {
            var type = typeof(T);

            var name = GetTableName(type);

            var sb = new StringBuilder(null);
            sb.AppendFormat("insert into {0} (", name);

            var allProperties = TypePropertiesCache(type);
            for (var i = 0; i < allProperties.Count(); i++)
            {
                var property = allProperties.ElementAt(i);
                if (property.IsKey && property.IsKeyIncrement) continue;
                if (property.IsIgnore && property.GetIgnore(IgnoreType.Insert)) continue;

                sb.Append(property.Column);
                sb.Append(", ");
            }
            sb.Remove(sb.Length - 2, 2);
            sb.Append(") values (");
            for (var i = 0; i < allProperties.Count(); i++)
            {
                var property = allProperties.ElementAt(i);
                if (property.IsKey && property.IsKeyIncrement) continue;
                if (property.IsIgnore && property.GetIgnore(IgnoreType.Insert)) continue;

                sb.AppendFormat("#{0}#", property.Name);
                sb.Append(", ");
            }
            sb.Remove(sb.Length - 2, 2);
            sb.Append(") ");
            var inserted = SqlMapper.Execute(connection, sb.ToString(), entityToInsert, transaction, commandTimeout, null);
            return inserted > 0;
        }

        /// <summary>
        /// Updates entity in table "Ts", checks if the entity is modified if the entity is tracked by the Get() extension.
        /// </summary>
        /// <typeparam name="T">Type to be updated</typeparam>
        /// <param name="connection">Open SqlConnection</param>
        /// <param name="entityToUpdate">Entity to be updated</param>
        /// <returns>true if updated, false if not found or not modified (tracked entities)</returns>
        public static bool Update<T>(this IDbConnection connection, T entityToUpdate, IDbTransaction transaction = null, int? commandTimeout = null) where T : class
        {
            var proxy = entityToUpdate as IProxy;
            if (proxy != null)
            {
                if (!proxy.IsDirty) return false;
            }

            var type = typeof(T);

            var allProperties = TypePropertiesCache(type);
            var keyProperties = allProperties.Where(op => op.IsKey);
            if (keyProperties.Count() == 0)
                throw new ArgumentException("Entity must have at least one [Key] property");

            var name = GetTableName(type);

            var sb = new StringBuilder();
            sb.AppendFormat("update {0} set ", name);

            for (var i = 0; i < allProperties.Count(); i++)
            {
                var property = allProperties.ElementAt(i);
                if (property.IsKey) continue;
                if (property.IsIgnore && property.GetIgnore(IgnoreType.Update)) continue;
                sb.AppendFormat("{0} = #{1}#", property.Column, property.Name);
                sb.AppendFormat(", ");
            }
            sb.Remove(sb.Length - 2, 2);
            sb.Append(" where ");
            for (var i = 0; i < keyProperties.Count(); i++)
            {
                var property = keyProperties.ElementAt(i);
                sb.AppendFormat("{0} = #{1}#", property.Column, property.Name);
                sb.AppendFormat(" and ");
            }
            sb.Remove(sb.Length - 5, 5);
            var updated = SqlMapper.Execute(connection, sb.ToString(), entityToUpdate, transaction, commandTimeout, null);
            return updated > 0;
        }

        /// <summary>
        /// Delete entity in table "Ts".
        /// </summary>
        /// <typeparam name="T">Type of entity</typeparam>
        /// <param name="connection">Open SqlConnection</param>
        /// <param name="entityToDelete">Entity to delete</param>
        /// <returns>true if deleted, false if not found</returns>
        public static bool Delete<T>(this IDbConnection connection, T entityToDelete, IDbTransaction transaction = null, int? commandTimeout = null) where T : class
        {
            var type = typeof(T);

            var keyProperties = TypePropertiesCache(type).Where(op => op.IsKey);
            if (keyProperties.Count() == 0)
                throw new ArgumentException("Entity must have at least one [Key] property");

            var name = GetTableName(type);

            var sb = new StringBuilder();
            sb.AppendFormat("delete from {0} where ", name);

            for (var i = 0; i < keyProperties.Count(); i++)
            {
                var property = keyProperties.ElementAt(i);
                sb.AppendFormat("{0} = #{1}#", property.Column, property.Name);
                sb.AppendFormat(" and ");
            }
            sb.Remove(sb.Length - 5, 5);
            var deleted = SqlMapper.Execute(connection, sb.ToString(), entityToDelete, transaction, commandTimeout, null);
            return deleted > 0;
        }


        class ProxyGenerator
        {
            private static readonly Dictionary<Type, object> TypeCache = new Dictionary<Type, object>();

            private static AssemblyBuilder GetAsmBuilder(string name)
            {
                var assemblyBuilder = Thread.GetDomain().DefineDynamicAssembly(new AssemblyName { Name = name },
                    AssemblyBuilderAccess.Run);       //NOTE: to save, use RunAndSave

                return assemblyBuilder;
            }

            public static T GetClassProxy<T>()
            {
                // A class proxy could be implemented if all properties are virtual
                //  otherwise there is a pretty dangerous case where internal actions will not update dirty tracking
                throw new NotImplementedException();
            }

            public static T GetInterfaceProxy<T>()
            {
                Type typeOfT = typeof(T);

                if (TypeCache.ContainsKey(typeOfT))
                {
                    return (T)TypeCache[typeOfT];
                }
                var assemblyBuilder = GetAsmBuilder(typeOfT.Name);

                var moduleBuilder = assemblyBuilder.DefineDynamicModule("SqlMapperExtensions." + typeOfT.Name); //NOTE: to save, add "asdasd.dll" parameter

                var interfaceType = typeof(Dapper.Contrib.Extensions.SqlMapperExtensions.IProxy);
                var typeBuilder = moduleBuilder.DefineType(typeOfT.Name + "_" + Guid.NewGuid(),
                    TypeAttributes.Public | TypeAttributes.Class);
                typeBuilder.AddInterfaceImplementation(typeOfT);
                typeBuilder.AddInterfaceImplementation(interfaceType);

                //create our _isDirty field, which implements IProxy
                var setIsDirtyMethod = CreateIsDirtyProperty(typeBuilder);

                // Generate a field for each property, which implements the T
                foreach (var property in typeof(T).GetProperties())
                {
                    var isId = property.GetCustomAttributes(true).Any(a => a is KeyAttribute);
                    CreateProperty<T>(typeBuilder, property.Name, property.PropertyType, setIsDirtyMethod, isId);
                }

                var generatedType = typeBuilder.CreateType();

                //assemblyBuilder.Save(name + ".dll");  //NOTE: to save, uncomment

                var generatedObject = Activator.CreateInstance(generatedType);

                TypeCache.Add(typeOfT, generatedObject);
                return (T)generatedObject;
            }

            private static MethodInfo CreateIsDirtyProperty(TypeBuilder typeBuilder)
            {
                var propType = typeof(bool);
                var field = typeBuilder.DefineField("_" + "IsDirty", propType, FieldAttributes.Private);
                var property = typeBuilder.DefineProperty("IsDirty",
                                               System.Reflection.PropertyAttributes.None,
                                               propType,
                                               new Type[] { propType });

                const MethodAttributes getSetAttr = MethodAttributes.Public | MethodAttributes.NewSlot | MethodAttributes.SpecialName |
                                                    MethodAttributes.Final | MethodAttributes.Virtual | MethodAttributes.HideBySig;

                // Define the "get" and "set" accessor methods
                var currGetPropMthdBldr = typeBuilder.DefineMethod("get_" + "IsDirty",
                                             getSetAttr,
                                             propType,
                                             Type.EmptyTypes);
                var currGetIL = currGetPropMthdBldr.GetILGenerator();
                currGetIL.Emit(OpCodes.Ldarg_0);
                currGetIL.Emit(OpCodes.Ldfld, field);
                currGetIL.Emit(OpCodes.Ret);
                var currSetPropMthdBldr = typeBuilder.DefineMethod("set_" + "IsDirty",
                                             getSetAttr,
                                             null,
                                             new Type[] { propType });
                var currSetIL = currSetPropMthdBldr.GetILGenerator();
                currSetIL.Emit(OpCodes.Ldarg_0);
                currSetIL.Emit(OpCodes.Ldarg_1);
                currSetIL.Emit(OpCodes.Stfld, field);
                currSetIL.Emit(OpCodes.Ret);

                property.SetGetMethod(currGetPropMthdBldr);
                property.SetSetMethod(currSetPropMthdBldr);
                var getMethod = typeof(Dapper.Contrib.Extensions.SqlMapperExtensions.IProxy).GetMethod("get_" + "IsDirty");
                var setMethod = typeof(Dapper.Contrib.Extensions.SqlMapperExtensions.IProxy).GetMethod("set_" + "IsDirty");
                typeBuilder.DefineMethodOverride(currGetPropMthdBldr, getMethod);
                typeBuilder.DefineMethodOverride(currSetPropMthdBldr, setMethod);

                return currSetPropMthdBldr;
            }

            private static void CreateProperty<T>(TypeBuilder typeBuilder, string propertyName, Type propType, MethodInfo setIsDirtyMethod, bool isIdentity)
            {
                //Define the field and the property 
                var field = typeBuilder.DefineField("_" + propertyName, propType, FieldAttributes.Private);
                var property = typeBuilder.DefineProperty(propertyName,
                                               System.Reflection.PropertyAttributes.None,
                                               propType,
                                               new Type[] { propType });

                const MethodAttributes getSetAttr = MethodAttributes.Public | MethodAttributes.Virtual |
                                                    MethodAttributes.HideBySig;

                // Define the "get" and "set" accessor methods
                var currGetPropMthdBldr = typeBuilder.DefineMethod("get_" + propertyName,
                                             getSetAttr,
                                             propType,
                                             Type.EmptyTypes);

                var currGetIL = currGetPropMthdBldr.GetILGenerator();
                currGetIL.Emit(OpCodes.Ldarg_0);
                currGetIL.Emit(OpCodes.Ldfld, field);
                currGetIL.Emit(OpCodes.Ret);

                var currSetPropMthdBldr = typeBuilder.DefineMethod("set_" + propertyName,
                                             getSetAttr,
                                             null,
                                             new Type[] { propType });

                //store value in private field and set the isdirty flag
                var currSetIL = currSetPropMthdBldr.GetILGenerator();
                currSetIL.Emit(OpCodes.Ldarg_0);
                currSetIL.Emit(OpCodes.Ldarg_1);
                currSetIL.Emit(OpCodes.Stfld, field);
                currSetIL.Emit(OpCodes.Ldarg_0);
                currSetIL.Emit(OpCodes.Ldc_I4_1);
                currSetIL.Emit(OpCodes.Call, setIsDirtyMethod);
                currSetIL.Emit(OpCodes.Ret);

                //TODO: Should copy all attributes defined by the interface?
                if (isIdentity)
                {
                    var keyAttribute = typeof(KeyAttribute);
                    var myConstructorInfo = keyAttribute.GetConstructor(new Type[] { });
                    var attributeBuilder = new CustomAttributeBuilder(myConstructorInfo, new object[] { });
                    property.SetCustomAttribute(attributeBuilder);
                }

                property.SetGetMethod(currGetPropMthdBldr);
                property.SetSetMethod(currSetPropMthdBldr);
                var getMethod = typeof(T).GetMethod("get_" + propertyName);
                var setMethod = typeof(T).GetMethod("set_" + propertyName);
                typeBuilder.DefineMethodOverride(currGetPropMthdBldr, getMethod);
                typeBuilder.DefineMethodOverride(currSetPropMthdBldr, setMethod);
            }

        }
    }




    public class MyMemberInfo : MemberInfo
    {
        //Column
        public string Column { get; set; }
        //Key
        public bool IsKey { get; set; }
        public bool IsKeyIncrement { get; set; }
        //Ignore
        public bool IsIgnore { get; set; }
        public IgnoreType IgnoreType { get; set; }

        //
        public PropertyInfo PropertyInfo { get; set; }
        public FieldInfo FieldInfo { get; set; }

        //扩展方法
        public bool GetIgnore(IgnoreType type)
        {
            if (IgnoreType == IgnoreType.All) return true;
            switch (type)
            {
                case IgnoreType.All:
                    return true;
                case IgnoreType.Select:
                    return (IgnoreType == IgnoreType.Select || IgnoreType == IgnoreType.SelectAndInsert || IgnoreType == IgnoreType.SelectAndUpdate);
                case IgnoreType.Insert:
                    return (IgnoreType == IgnoreType.Insert || IgnoreType == IgnoreType.SelectAndInsert || IgnoreType == IgnoreType.InsertAndUpdate);
                case IgnoreType.Update:
                    return (IgnoreType == IgnoreType.Update || IgnoreType == IgnoreType.SelectAndUpdate || IgnoreType == IgnoreType.InsertAndUpdate);
            }
            return false;
        }
        public void SetValue(object obj, object value)
        {
            if (PropertyInfo != null)
                PropertyInfo.SetValue(obj, value, null);
            else if (FieldInfo != null)
                FieldInfo.SetValue(obj, value);
        }


        //重写抽象类
        public override Type DeclaringType
        {
            get
            {
                if (PropertyInfo != null)
                    return PropertyInfo.DeclaringType;
                else if (FieldInfo != null)
                    return FieldInfo.DeclaringType;
                else
                    return null;
            }
        }

        public override object[] GetCustomAttributes(Type attributeType, bool inherit)
        {
            if (PropertyInfo != null)
                return PropertyInfo.GetCustomAttributes(attributeType, inherit);
            else if (FieldInfo != null)
                return FieldInfo.GetCustomAttributes(attributeType, inherit);
            else
                return null;
        }

        public override object[] GetCustomAttributes(bool inherit)
        {
            if (PropertyInfo != null)
                return PropertyInfo.GetCustomAttributes(inherit);
            else if (FieldInfo != null)
                return FieldInfo.GetCustomAttributes(inherit);
            else
                return null;
        }

        public override bool IsDefined(Type attributeType, bool inherit)
        {
            if (PropertyInfo != null)
                return PropertyInfo.IsDefined(attributeType, inherit);
            else if (FieldInfo != null)
                return FieldInfo.IsDefined(attributeType, inherit);
            else
                return false;
        }

        public override MemberTypes MemberType
        {
            get
            {
                if (PropertyInfo != null)
                    return PropertyInfo.MemberType;
                else if (FieldInfo != null)
                    return FieldInfo.MemberType;
                else
                    return MemberTypes.All;
            }
        }

        public override string Name
        {
            get
            {
                if (PropertyInfo != null)
                    return PropertyInfo.Name;
                else if (FieldInfo != null)
                    return FieldInfo.Name;
                else
                    return null;
            }
        }

        public override Type ReflectedType
        {
            get
            {
                if (PropertyInfo != null)
                    return PropertyInfo.ReflectedType;
                else if (FieldInfo != null)
                    return FieldInfo.ReflectedType;
                else
                    return null;
            }
        }
    }
    [AttributeUsage(AttributeTargets.Class)]
    public class TableAttribute : Attribute
    {
        public TableAttribute(string tableName)
        {
            Name = tableName;
        }
        public string Name { get; set; }
    }
    [AttributeUsage(AttributeTargets.Class)]
    public class ColumnAttribute : Attribute
    {
        public ColumnAttribute(string columnName)
        {
            Name = columnName;
        }
        public string Name { get; set; }
    }
    [AttributeUsage(AttributeTargets.Property)]
    public class KeyAttribute : Attribute
    {
        public KeyAttribute()
        {
            IsIncrement = false;
        }
        public KeyAttribute(bool isIncrement)
        {
            IsIncrement = isIncrement;
        }
        public bool IsIncrement { get; set; }
    }
    [AttributeUsage(AttributeTargets.Property)]
    public class IgnoreAttribute : Attribute
    {
        public IgnoreAttribute()
        {
            Type = IgnoreType.All;
        }
        public IgnoreAttribute(IgnoreType type)
        {
            Type = type;
        }
        public IgnoreType Type { get; set; }
    }
    public enum IgnoreType
    {
        All,
        Select,
        Insert,
        Update,
        SelectAndInsert,
        SelectAndUpdate,
        InsertAndUpdate
    }
}
